///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Filename:		Pre-emptive Ping Pong Overlay Manager v1.asm
//							(c) ANALOG DEVICES 2001, gjf
//
// Date:				10/29/2001
// Purpose:			Software Overlay Manager Function
//
// Inputs:			primary ax0 (PLIT_SYMBOL_OVERLAYID), primary ay0 (PLIT_SYMBOL_ADDRESS)
// Outputs:			primary i7 (used to call overlay module for execution of the overlay)
//
// Description:	This progam implements a simple software overlay manager which is uses the DSP's DMA engine to load
//							in the overlay (code or data) from external memory. This code module has been left generic enough to
//							allow for simple code reuse and modifications.
//							
// Registers Used:
//
//							primary register set:		ax0, ay0, sr2, sr1, sr0, ar, dmpg2, iopg, IJPG (not saved or restored!!!!)
//							primary DAGs:						i7, m7
//							secondary register set:	none
//							secondary DAGs:					none
//
// Additional Information:
//
//							The previous version of this overlay manager utilized a software stack (stored in DM) to perform the
//							context save-restore of ALL of the core registers. This may prove to be impractical for some systems
//							due to the additional overhead of the context save-restore of all of the 2191's registers. This version
//							consumes only a small portion of the 2191's register set which allows for fast context saving-restoring
//							of the registers. Also, since this overlay manager does not use the secondary register set, this code
//							module can be interrupted without any adverse side-effects.
//
//							Currently there is a bug in the linker which places the overlay ID# as the first location (target 'live'
//							address) of the overlay. This affects the count and actual the live address of the overlay. Because of
//							this some additional code has been implemented (labeled 'bug' in the commented section of the code) as
//							a workaround for this bug.
//
// Revision Information:
//
//							rev 0.0 - (10/23/01) -	This revision assumes that ALL calls to the loaded overlay modules are within internal
//																			PM memory. Therefore the IJPG register is NEVER saved or restored. If the user wishes
//																			to execute PM overlays that 'run' from external memory, then care must be taken to
//																			ensure that the IJPG register is properly saved and restored on the overlay stack.
//
//							rev 1.0 - (10/24/01) -	Added support for DM overlays. In order to generically support PM and DM overlays
//																			this version of the overlay manager MUST configure the EMI_CTRL and E_STAT registers
//																			accordingly to support the loading in of the PM and DM overlays from external memory.
//																			Again, this overlay manager assumes that the 'live' addresses of both the PM and DM
//																			overlays reside in internal memory of the 2191.
//
//							rev 1.0 - (10/24/01) -	This revision does not use any of the linker generated overlay constants. Instead the
//																			user must hand build the overlay arrays (located in the file "DM Data.asm") to
//																			store the run and live size of the overlay modules. This will be cumbersome for users
//																			who have a significant amount of overlay modules in their project.
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Here is the total number of overlay modules in our project
#define TOTAL_NUM_OF_OVLYS 7

// Here is the total number of PM overlay modules in our project
#define TOTAL_NUM_OF_PM_OVLYS 4

#include "2191asm.h"
.section/dm	seg_data;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DMA descriptor blocks below. One descriptor is used for the DMA read channel, the other for the DMA write channel.
// We'll initialize these blocks to zero just as a precaution.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

.var WR_DMA_DESC_BLOCK [5] =	0x0000,		// write DMA configuration register
															0x0000,		// write DMA start page (internal DSP 'live' address of overlay module (ALWAYS assumed as zero!!))
															0x0000,		// write DMA start address (internal DSP 'run' address of overlay module)
															0x0000,		// write DMA word count
															0x0000;		// write DMA next descriptor pointer
															
.var RD_DMA_DESC_BLOCK [5] =  0x0000,		// read DMA configuration register
															0x0000,		// read DMA start page
															0x0000,		// read DMA start address
															0x0000,		// read DMA word count
															0x0000;		// read DMA next descriptor pointer

.var end_dma = 0x0000;									// this variable is used to terminate the DMA (set this variable to a non-zero value for chained DMAs)

.var run_space_var;											// used to determine which overlay run space the overlay we requested resides in
//.var pm_ovl1_okay = 0;									// used by overlay functions to return to proper address after overlay function call
//.var pm_ovl2_okay = 0;									// used by overlay functions to return to proper address after overlay function call

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// This array below is used to keep track of the number of times each overlay module has been called in order to
// determine which overlay module is used more frequently by the overlay manager.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

.var ovly_module_fetch_count[TOTAL_NUM_OF_OVLYS] = 0, 0, 0, 0, 0, 0, 0;

.global WR_DMA_DESC_BLOCK;
.global RD_DMA_DESC_BLOCK;
.global end_dma;
.global run_space_var;
//.global pm_ovl1_okay;
//.global pm_ovl2_okay;
.global ovly_module_fetch_count;


.section/pm	seg_code;

.global _OverlayManager;								// address label for the overlay manager function (which is called from the main program)

//////////////////////////////////////////////////////////////////////////////////////////////////
// The runAddresses for the overlay modules are not needed here since they are defined at runtime
//////////////////////////////////////////////////////////////////////////////////////////////////

.extern liveAddresses;									// the 'live' address for the overlay (external memory location)
.extern runWordSize;										// number of internal DSP (PM or DM) memory locations the overlay will occupy
.extern liveWordSize;										// number of external (PM or DM) memory locations the overlay will occupy (same as runWordSize for 2191)

.extern ov_stack;												// software runtime stack (in DM) used for context save-restore by overlay manager function

.extern save_ax0;												// temporary memory locations for saving-restoring of ax0 register
.extern save_ay0;												// temporary memory locations for saving-restoring of ay0 register

.extern ovly_ID_temp;										// variable used to temporarily store the ID# of the overlay

.extern curr_PM_ovly_ID1;								// variable used to check if the PM overlay region is already loaded into internal PM run memory space #1
.extern curr_PM_ovly_ID2;								// variable used to check if the PM overlay region is already loaded into internal PM run memory space #2
.extern curr_DM_ovly_ID1;								// variable used to check if the DM overlay region is already loaded into internal DM run memory space #1
.extern curr_DM_ovly_ID2;								// variable used to check if the DM overlay region is already loaded into internal DM run memory space #2

.extern LRU_PM_Ovly_1;									// LRU variable used to define if internal PM run memory space #1 is least recently used or not
.extern LRU_PM_Ovly_2;									// LRU variable used to define if internal PM run memory space #2 is least recently used or not

.extern LRU_DM_Ovly_1;									// LRU variable used to define if internal DM run memory space #1 is least recently used or not
.extern LRU_DM_Ovly_2;									// LRU variable used to define if internal DM run memory space #2 is least recently used or not



_OverlayManager:

	nop;																	// NOP added here for debugging purposes only

Context_Save:

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Save registers that the overlay manager will be using onto the overlay manager stack.
	// (This will come in handy when we'll be pre-emptively calling an overlay.)
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = dmpg2;
	dm(ov_stack+1) = ax0;									// save dmpg2 register onto overlay manager stack (push)
	dm(ov_stack+2) = ar;									// save ar register onto overlay manager stack (push)
	dm(ov_stack+3) = m7;									// save m7 register onto overlay manager stack (push)
	dm(ov_stack+4) = i7;									// save i7 register onto overlay manager stack (push)
	dm(ov_stack+5) = sr0;									// save sr0 register onto overlay manager stack (push)
	dm(ov_stack+6) = sr1;									// save sr1 register onto overlay manager stack (push)
	dm(ov_stack+7) = sr2;									// save sr2 register onto overlay manager stack (push)
	ax0 = iopg;
	dm(ov_stack+8) = ax0;									// save iopg register onto overlay manager stack (push)
	dm(ov_stack+9) = ay0;									// save ay0 register onto overlay manager stack (push)

	dm(ov_stack+10) = i0;									// save i0 register onto overlay manager stack (push)
	dm(ov_stack+11) = m0;									// save m0 register onto overlay manager stack (push)
	dm(ov_stack+12) = m1;									// save m1 register onto overlay manager stack (push)
	dm(ov_stack+13) = m2;									// save m2 register onto overlay manager stack (push)

Get_Overlay_ID_value:

	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here we check to see if the overlay we intend to fetch is already loaded into one of the overlay 'run' spaces.
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ar = dm(ov_stack);										// get overlay number of the overlay that we want to fetch
	ar = ar - 1;													// decrement overlay number to get proper overlay ID# (off-by-one offset)
	ax0 = ar;															// place overlay ID# into ax0 for comparison
	dm(ov_stack) = ar;										// return the overlay ID# to overlay stack (needed for EMI DMA complete ISR)


Update_Overlay_Fetch_Count:

	// new for LFU
	i0 = ovly_module_fetch_count;					// i0 points to the head of the overlay fetch count array
	m0 = dm(ov_stack);										// needed for overlay ID# offset into overlay fetch count array
	ar = dm(i0 + m0);											// fetch the overlay fetch count value for this overlay ID#
	ar = ar + 1;													// increment the overlay fetch count value
	dm(i0 + m0) = ar;											// update the overlay fetch count value in the array (we'll need to do this instruction anyway)
	ar = tstbit 14 of ar;									// check to see if the count value is getting appreciably large
	if eq jump Is_Ovly_Loaded_Already;		// if the count is less than 16k then we're okay, otherwise scale down all of the overlay count values below
	
	
Scale_Down_Count:

	//////////////////////////////////////////////////////////////////////////////
	// Here we need to scale down all of the count values in our array by a
	// pre-determined value to ensure that our LFU comparisons don't blow up if
	// one of the count values exceeds 16k, since we don't want to do a compare
	// with one of the values being negative!!
	//////////////////////////////////////////////////////////////////////////////
	
	m1 = 0;																// assign m1 to zero so we don't stride through our buffer
	m2 = 1;																// assign m1 to one so we do stride through our buffer
	
	cntr = TOTAL_NUM_OF_OVLYS;						// assign counter a value of the total number of overlays in our overlay count array
	do Kill_Overflow until ce;						// start of loop body
	
		ar = dm(i0, m1);										// read element from overlay count array (don't update index register yet)
		sr = lshift ar by -1 (hi);					// scale down the overlay count element value by 2 (this can be increased if needed)

Kill_Overflow:

		dm(i0, m2) = sr1;										// place the scaled down overlay count value back into its proper array element (end of loop)


Restore_M1_Reg:

	m2 = dm(ov_stack+13);									// restore the M2 register from the stack (pop)
	m1 = dm(ov_stack+12);									// restore the M1 register from the stack (pop)
	

Is_Ovly_Loaded_Already:

	ay0 = dm(curr_PM_ovly_ID1);						// get ID# of overlay currently in PM run space #1
	ar = ax0 - ay0;												// see if the overlay we are fetching is already loaded into PM run space #1
	if eq jump Goto_PM_Run_1;							// if we have a hit then just go to routine to jump to PM overlay in PM run space #1

	ay0 = dm(curr_PM_ovly_ID2);						// get ID# of overlay currently in PM run space #2
	ar = ax0 - ay0;												// see if the overlay we are fetching is already loaded into PM run space #1
	if eq jump Goto_PM_Run_2;							// if we have a hit then just go to routine to jump to PM overlay in PM run space #1
	
	ay0 = dm(curr_DM_ovly_ID1);						// get ID# of overlay currently in DM run space #1
	ar = ax0 - ay0;												// see if the overlay we are fetching is already loaded into DM run space #1
	if eq jump DM_Context_Restore;				// if we have a hit then just exit the overlay manager and return back to main program
	
	ay0 = dm(curr_DM_ovly_ID2);						// get ID# of overlay currently in DM run space #2
	ar = ax0 - ay0;												// see if the overlay we are fetching is already loaded into DM run space #1
	if eq jump DM_Context_Restore;				// if we have a hit then just exit the overlay manager and return back to main program
	
	nop;																	// NOP added here for debugging purpose only
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////////
	// If we have reached the NOP instruction above, then that means that the overlay module that we are
	// fetching has not already been loaded into internal PM or DM memory. Therefore, we will just continue
	// on to find out whether the overlay ID# is for a PM or DM run space, and then determine which run space
	// we need to DMA the overlay module into starting below...
	///////////////////////////////////////////////////////////////////////////////////////////////////////////

Find_Overlay:

	///////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here is where we determine the ID# of the overlay that we wish to fetch to DMA it into one of the two
	// run spaces for the PM or DM overlay module.
	///////////////////////////////////////////////////////////////////////////////////////////////////////////

	dmpg2 = 0;																	// switch to the internal memory page #0 since all of the overlay symbols reside in internal DM
	m7 = dm(ov_stack);													// m7 is the offset needed to get _ov values from the "liveaddresses", "runaddresses", and "pageaddresses" buffers
	dm(ovly_ID_temp) = m7;											// save off the ID# of the overlay to a temp variable
	
	ay0 = TOTAL_NUM_OF_PM_OVLYS;								// ay0 = total # of PM overlay modules in the project
	ar = ax0 - ay0;															// subtract total # of PM overlays from overlay ID # that we are fetching to determine if overlay we are to load is PM or DM
//	if le jump Check_PM_LRU;							// if ID# we want is for PM then jump to see what run space we can load this overlay into
	// new for LFU
	if le jump Check_PM_Overlay_First_Time;			// if the ID# we want is for PM then jump to see which PM overlay live space has the more commonly used overlay module


Check_DM_Overlay_First_Time:

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here we check to see if we're executing the overlay manager for the first time. We need to do this here because we
	// initialize the two DM variables 'curr_DM_ovly_ID1' and 'curr_DM_ovly_ID2' to a value of 0xffff. This would totally hose
	// us here if we used these overlay ID values for offsets into our overlay fetch count array!!!!!!!!!
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = dm(curr_DM_ovly_ID1);									// ax0 equals the overlay ID# of the overlay that's currently in DM 'live' space #1
	ay0 = dm(curr_DM_ovly_ID2);									// ay0 equals the overlay ID# of the overlay that's currently in DM 'live' space #2
	ar = ax0 - ay0;															// compare the two values (they should only be equal the first time we execute the overlay manager
	if eq jump Load_DM_Ovly_Run_1;							// we know at this point that both run spaces are empty, so we'll load the 1st one here


Compare_DM_Ovly_Count:

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here we check the count value for each overlay module to see which module to replace with the new overlay which we
	// currently are requesting. We simply compare the fetch count values for the two overlay modules that reside in their
	// respective overlay live spaces, and replace the overlay module which is used less frequently with the current overlay
	// that we're fetching.
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	m0 = dm(curr_DM_ovly_ID1);									// m0 equals the overlay ID# of the overlay that's currently in PM 'live' space #1
	ax0 = dm(i0 + m0);													// ax0 equals the number of times this overlay module has been called
	m0 = dm(curr_DM_ovly_ID2);									// m0 equals the overlay ID# of the overlay that's currently in PM 'live' space #2
	ay0 = dm(i0 + m0);													// ay0 equals the number of times this overlay module has been called
	ar = ax0 - ay0;															// subtract the two values to see which one has been called more times
	
	if lt jump Load_DM_Ovly_Run_1;							// if overlay in live space #1 called less times than the overlay in live space #2 then replace live space #1
	if gt jump Load_DM_Ovly_Run_2;							// if overlay in live space #2 called less times than the overlay in live space #1 then replace live space #2

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// If both overlay count values are equal, then we 'fall through' to the code below where we simply check the value of the
	// LRU variable for each overlay run segment, thereby replacing the oldest overlay module.
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	

Check_DM_LRU:

	///////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here is where we check the DM LRU variable for DM overlay run spaces #1 and #2 to determine which DM
	// overlay run space (#1 or #2) we need to load the overlay module we wish to fetch.
	///////////////////////////////////////////////////////////////////////////////////////////////////////////

	ar = dm(LRU_DM_Ovly_1);									// load in the LRU variable for DM Overlay run space #1
	ar = ar - 1;														// test to see if the LRU variable is set to 1
	if eq jump Load_DM_Ovly_Run_1;					// if LRU variable is 1 for this run space then load the overlay we want into that run space


Load_DM_Ovly_Run_2:

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here is where we initialize a portion of the DMA descriptor block to load the overlay module into internal
	// DM overlay run segment #2 (DM 0xc000 - 0xffff).
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = 0xc000;														// assign 1st location of DM overlay run space #2 into ax0
	dm(WR_DMA_DESC_BLOCK+2) = ax0;					// write the overlay 'run' address for MEM DMA WR address reg into its location in MEM DMA WR descriptor block
	
	ax0 = 0x0000;														// set ax0 to zero
	dm(LRU_DM_Ovly_2) = ax0;								// clear the LRU variable for DM overlay run space #2 (set to MRU value)
	ax0 = 0x0001;														// set ax0 to 1
	dm(LRU_DM_Ovly_1) = ax0;								// set the LRU variable for DM overlay run space #1 since it now is LRU
	
	jump Load_DM_Ovly (db);									// delayed branch jump to "Load_DM_Ovly" section of this overlay manager
	ar = dm(ovly_ID_temp);									// read the temp value of the overlay we are fetching into ar register
	dm(curr_DM_ovly_ID2) = ar;							// write value of overlay we are fetching into current ID variable for run space #2


Load_DM_Ovly_Run_1:

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here is where we initialize a portion of the DMA descriptor block to load the overlay module into internal
	// DM overlay run segment #1 (DM 0x9000 - 0xbfff).
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = 0x9000;														// assign 1st location of DM overlay run space #1 into ax0
	dm(WR_DMA_DESC_BLOCK+2) = ax0;					// write the overlay 'run' address for MEM DMA WR address reg into its location in MEM DMA WR descriptor block
	
	ax0 = 0x0000;														// set ax0 to zero
	dm(LRU_DM_Ovly_1) = ax0;								// clear the LRU variable for DM overlay run space #1 (set to MRU value)
	ax0 = 0x0001;														// set ax0 to 1
	dm(LRU_DM_Ovly_2) = ax0;								// set the LRU variable for DM overlay run space #2 since it now is LRU
	
	ar = dm(ovly_ID_temp);									// read the temp value of the overlay we are fetching into ar register
	dm(curr_DM_ovly_ID1) = ar;							// write value of overlay we are fetching into current ID variable for run space #1

	
Load_DM_Ovly:

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here is where we need to configure the EMI for the loading of the external DM overlay. Keep in mind that for this example
	// since we are booting from 8-bit external memory, we also must load in the overlays from 8-bit external memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	iopg = External_Memory_Interface_Page;			// switch to the EMI interface page (0x06)
	ax0 = 0x0030;																// assign ax0 to write appropriate value to EMI Control Register
	io(EMICTL) = ax0;														// configure EMI for 8-bit ext bus, ~RD/~WR acive low, DMA split disabled

	iopg = External_Access_Bridge_Page; 				// set to IOPG 0x00
	jump Parse_Live_Address (db);								// continue on with the remainder of the overlay manager code
	ax0 = 0x0000;
	io(E_STAT) = ax0;														// configure External Access Bridge for 16-bit external data accesses


Check_PM_Overlay_First_Time:

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here we check to see if we're executing the overlay manager for the first time. We need to do this here because we
	// initialize the two PM variables 'curr_PM_ovly_ID1' and 'curr_PM_ovly_ID2' to a value of 0xffff. This would totally hose
	// us here if we used these overlay ID values for offsets into our overlay fetch count array!!!!!!!!!
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = dm(curr_PM_ovly_ID1);									// ax0 equals the overlay ID# of the overlay that's currently in PM 'live' space #1
	ay0 = dm(curr_PM_ovly_ID2);									// ay0 equals the overlay ID# of the overlay that's currently in PM 'live' space #2
	ar = ax0 - ay0;															// compare the two values (they should only be equal the first time we execute the overlay manager
	if eq jump Load_PM_Ovly_Run_1;							// we know at this point that both run spaces are empty, so we'll load the 1st one here


Compare_PM_Ovly_Count:

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here we check the count value for each overlay module to see which module to replace with the new overlay which we
	// currently are requesting. We simply compare the fetch count values for the two overlay modules that reside in their
	// respective overlay live spaces, and replace the overlay module which is used less frequently with the current overlay
	// that we're fetching.
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	m0 = dm(curr_PM_ovly_ID1);									// m0 equals the overlay ID# of the overlay that's currently in PM 'live' space #1
	ax0 = dm(i0 + m0);													// ax0 equals the number of times this overlay module has been called
	m0 = dm(curr_PM_ovly_ID2);									// m0 equals the overlay ID# of the overlay that's currently in PM 'live' space #2
	ay0 = dm(i0 + m0);													// ay0 equals the number of times this overlay module has been called
	ar = ax0 - ay0;															// subtract the two values to see which one has been called more times
	
	if lt jump Load_PM_Ovly_Run_1;							// if overlay in live space #1 called less times than the overlay in live space #2 then replace live space #1
	if gt jump Load_PM_Ovly_Run_2;							// if overlay in live space #2 called less times than the overlay in live space #1 then replace live space #2

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// If both overlay count values are equal, then we 'fall through' to the code below where we simply check the value of the
	// LRU variable for each overlay run segment, thereby replacing the oldest overlay module.
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	

Check_PM_LRU:

	///////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here is where we check the PM LRU variable for PM overlay run spaces #1 and #2 to determine which PM
	// overlay run space (#1 or #2) we need to load the overlay module we wish to fetch.
	///////////////////////////////////////////////////////////////////////////////////////////////////////////

	ar = dm(LRU_PM_Ovly_1);									// load in the LRU variable for PM Overlay run space #1
	ar = ar - 1;														// test to see if the LRU variable is set to 1
	if eq jump Load_PM_Ovly_Run_1;					// if LRU variable is 1 for this run space then load the overlay we want into that run space


Load_PM_Ovly_Run_2:

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here is where we initialize a portion of the DMA descriptor block to load the overlay module into internal
	// PM overlay run segment #2 (PM 0x4000 - 0x7fff).
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = 0x4000;														// assign 1st location of PM overlay run space #2 into ax0
	dm(WR_DMA_DESC_BLOCK+2) = ax0;					// write the overlay 'run' address for MEM DMA WR address reg into its location in MEM DMA WR descriptor block
	dm(ov_stack+19) = ax0;									// save the overlay 'run' address onto the overlay stack (needed for target address for MEM DMA ISR)
	
	ax0 = 0x0000;														// set ax0 to zero
	dm(LRU_PM_Ovly_2) = ax0;								// clear the LRU variable for PM overlay run space #2 (set to MRU value)
	ax0 = 0x0001;														// set ax0 to 1
	dm(LRU_PM_Ovly_1) = ax0;								// set the LRU variable for PM overlay run space #1 since it now is LRU
	
	jump Load_PM_Ovly (db);									// delayed branch jump to "Load_DM_Ovly" section of this overlay manager
	ar = dm(ovly_ID_temp);									// read the temp value of the overlay we are fetching into ar register
	dm(curr_PM_ovly_ID2) = ar;							// write value of overlay we are fetching into current ID variable for run space #2


Load_PM_Ovly_Run_1:

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here is where we initialize a portion of the DMA descriptor block to load the overlay module into internal
	// PM overlay run segment #1 (PM 0x0400 - 0x3fff).
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = 0x0400;														// assign 1st location of PM overlay run space #1 into ax0
	dm(WR_DMA_DESC_BLOCK+2) = ax0;					// write the overlay 'run' address for MEM DMA WR address reg into its location in MEM DMA WR descriptor block
	dm(ov_stack+19) = ax0;									// save the overlay 'run' address onto the overlay stack (needed for target address for MEM DMA ISR)
	
	ax0 = 0x0000;														// set ax0 to zero
	dm(LRU_PM_Ovly_1) = ax0;								// clear the LRU variable for PM overlay run space #1 (set to MRU value)
	ax0 = 0x0001;														// set ax0 to 1
	dm(LRU_PM_Ovly_2) = ax0;								// set the LRU variable for PM overlay run space #2 since it now is LRU
	
	ar = dm(ovly_ID_temp);									// read the temp value of the overlay we are fetching into ar register
	dm(curr_PM_ovly_ID1) = ar;							// write value of overlay we are fetching into current ID variable for run space #1


Load_PM_Ovly:

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Here is where we need to configure the EMI for the loading of the external PM overlay. Keep in mind that for this example
// since we are booting from 8-bit external memory, we also must load in the overlays from 8-bit external memory
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	iopg = External_Memory_Interface_Page;			// switch to the EMI interface page (0x06)
	ax0 = 0x0030;																// configure EMI for 8-bit ext bus, ~RD/~WR acive low, DMA split disabled
	io(EMICTL) = ax0;

	iopg = External_Access_Bridge_Page; 				// set to IOPG 0x00
	ax0 = 0x0008;
	io(E_STAT) = ax0;														// configure External Access Bridge for 24-bit external data accesses


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// We don't need to update the DMA WR page register since it's initialized to zero in the descriptor block by default
// and also because for our example ALL of the overlay regions run from internal PM (i.e. ijpg=0).
// The code for implementing this will be provided below for additional modification of this example by the user.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	

Parse_Live_Address:

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Since the live address is a 24-bit value we have to parse this 24-bit value
// into the 8-bit page and 16-bit address offset fields to update the DMPGx and DAG registers accordingly
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	i7 = runWordSize;											// i7 points to head of the 'runWordSize' buffer
	ax0 = dm(i7 + m7);  									// get the word count of the desired overlay (from m7 offset)
	ar = ax0;															// ax0 equals the word count of the desired overlay
	dm(WR_DMA_DESC_BLOCK+3) = ar;					// write word count for MEM DMA word count into its location in the DMA WR descriptor block
	dm(RD_DMA_DESC_BLOCK+3) = ar;					// read word count for MEM DMA word count into its location in the DMA RD descriptor block

	i7 = liveAddresses;										// i7 points to head of liveAddresses buffer
	ar = pm(i7 + m7);											// get upper 16 MSBs of 24-bit address constant label of the desired overlay (from m7 offset)
	sr = lshift ar by 0x8 (lo);						// shift value 8 bits left to place page value in SR1
	ar = px;															// get 8 LSBs of 24-bit address constant label
	sr = sr or lshift ar by 0x0 (lo);			// OR the 8 LSBs with the parsed 16 LSBs from prior shift instruction
	i7 = sr0;															// i7 points to the live address of the desired overlay
	ar = i7;															// load target address into ar (bug)

	dm(RD_DMA_DESC_BLOCK+2) = ar;					// write overlay 'live' address for MEM DMA RD address reg into its location in MEM DMA RD descriptor block
	dm(RD_DMA_DESC_BLOCK+1) = sr1;				// write overlay 'live' page for MEM DMA RD page reg into its location in MEM DMA RD descriptor block

	ax0 = 0x800f;													// configure MEM DMA WR Config Reg for 24-bit writes (to int mem), enable DMA, DMA descriptor ownership, interrupt on completion
	dm(WR_DMA_DESC_BLOCK) = ax0;					// write this value for MEM DMA WR Config Reg into its location in MEM DMA WR descriptor block
	ax0 = 0x800d;													// configure MEM DMA RD Config Reg for 24-bit reads (from ext mem), enable DMA, DMA descriptor ownership, interrupt on completion
	dm(RD_DMA_DESC_BLOCK) = ax0;					// write this value for MEM DMA RD Config Reg into its location in MEM DMA RD descriptor block
	
	ax0 = end_dma;												// ax0 equals address of 'end_dma' variable
	dm(WR_DMA_DESC_BLOCK+4) = ax0;				// this writes the value of 'end_dma' to end the DMA chaining sequence (=0)
	dm(RD_DMA_DESC_BLOCK+4) = ax0;				// this writes the value of 'end_dma' to end the DMA chaining sequence (=0)
	

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Here's where we finally kick off the DMA and fetch the overlay module !!!!!!!!!
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	iopg =  Memory_DMA_Controller_Page;		// switch to MEM DMA IO page (0x01)

	ax0 = WR_DMA_DESC_BLOCK;							// ax0 points to the head of the write DMA descriptor block (linked-list)
	io(DMACW_CP) = ax0;										// load the address of the DMA write channel descriptor block into write chain pointer register
	ax0 = RD_DMA_DESC_BLOCK;							// ax0 points to the head of the write DMA descriptor block (linked-list)
	io(DMACR_CP) = ax0;										// load the address of the DMA read channel descriptor block into read chain pointer register

	ax0 = 0x1;
	io(DMACW_CPR) = ax0;									// set the descriptor ready bit for the DMA write channel
	io(DMACR_CPR) = ax0;									// set the descriptor ready bit for the DMA read channel

	io(DMACW_CFG) = ax0;									// enable the write DMA channel
	io(DMACR_CFG) = ax0;									// enable the read DMA channel

	ena int;															// enable global interrupts
	
	idle;																	// sit here and wait for the DMA interrupt to complete

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// At this point we're done with loading in the overlay region. Now all we have to do is restore the scratch registers used
// by the overlay manager and jump to the overlay function to begin execution.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Context_Restore:

	ay0 = dm(ov_stack+9);									// restore ay0 register (pop)
	ax0 = dm(ov_stack+8);									// restore iopg register value into ax0 register (pop)
	iopg = ax0;														// restore iopg register
	sr2 = dm(ov_stack+7);									// restore primary sr2 register (pop)
	sr1 = dm(ov_stack+6);									// restore primary sr1 register (pop)
	sr0 = dm(ov_stack+5);									// restore primary sr0 register (pop)
	i7 = dm(ov_stack+4);									// restore primary i7 register (pop)
	m7 = dm(ov_stack+3);									// restore primary m7 register (pop)
	ar = dm(ov_stack+2);									// restore primary ar register (pop)

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// we don't need to restore ax0 here because it's saved by the calling function (main program) already
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	flush cache;													// we don't want to cache any of the overlay manager instructions (self-modifying code & cache = bad!!)


Jump_To_Caller:

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Here we need to figure out if the DMA'ed overlay was for PM or DM. If it's for DM then we just execute an RTS and go back to
// the main program. If the overlay was for PM then we need to manipulate the PC stack. We need to do this manipulation of the
// stack because we call the overlay manager and we jump to the PM overlay function. Because of this we need to modify the stack
// so that we return to the correct address in main after executing the RTS instruction in the overlay function.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = dm(ov_stack);											// get the overlay ID# from the software stack (this value was saved previously by the overlay manager)
	ay0 = TOTAL_NUM_OF_PM_OVLYS;						// let ay0 equal the total number of PM overlays in our project
	ar = ax0 - ay0;													// determine if the overlay module that was DMAed into internal memory was for PM or DM
	
	if lt jump Goto_Target;									// if it's a PM overlay then we need to jump to its address
	rts (db);																// otherwise it's a DM overlay so return back to main program (and enable primary registers again)
	ax0 = dm(ov_stack+1); 								// restore dmpg2 reg value into ax0 (pop)
	dmpg2 = ax0;													// restore DAG dmpage 2 register

Goto_Target:
	
	dm(ov_stack+18) = i7;										// context save i7 register
	i7 = dm(ov_stack+19);										// i7 equals overlay target address
	jump (i7) (db);													// jump to the target address of the overlay (the RTS in overlay will restore the PC stack)
	i7 = dm(ov_stack+18);										// context restore i7 register
	nop;

Goto_PM_Run_1:

	/////////////////////////////////////////////////////////////////////////////////////////////
	// Here at this point in the overlay manager we have determined that the PM overlay function
	// that we are requesting already resides within it's proper overlay "run" space. Therefore
	// we will just perform a context restoration of the core registers and jump to the proper
	// PM overlay function.
	/////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = dm(ov_stack+8);									// restore iopg register value into ax0 register (pop)
	iopg = ax0;														// restore iopg register
	sr2 = dm(ov_stack+7);									// restore primary sr2 register (pop)
	sr1 = dm(ov_stack+6);									// restore primary sr1 register (pop)
	sr0 = dm(ov_stack+5);									// restore primary sr0 register (pop)
	m7 = dm(ov_stack+3);									// restore primary m7 register (pop)
	ax0 = dm(ov_stack+1); 								// restore dmpg2 reg value into ax0 (pop)
	dmpg2 = ax0;													// restore DAG dmpage 2 register
	ay0 = dm(ov_stack+9);									// restore ay0 register (pop)

	flush cache;													// we don't want to cache any of the overlay manager instructions (self-modifying code & cache = bad!!)

	i7 = 0x0400;													// load i7 with address of 1st location of PM run space #1 (PM 0x0400)
	jump (i7) (db);												// jump to location that i7 points to	
	ar = dm(ov_stack+2);									// restore primary ar register (pop)
	i7 = dm(ov_stack+4);									// restore primary i7 register (pop)


Goto_PM_Run_2:

	/////////////////////////////////////////////////////////////////////////////////////////////
	// Here at this point in the overlay manager we have determined that the PM overlay function
	// that we are requesting already resides within it's proper overlay "run" space. Therefore
	// we will just perform a context restoration of the core registers and jump to the proper
	// PM overlay function.
	/////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = dm(ov_stack+8);									// restore iopg register value into ax0 register (pop)
	iopg = ax0;														// restore iopg register
	sr2 = dm(ov_stack+7);									// restore primary sr2 register (pop)
	sr1 = dm(ov_stack+6);									// restore primary sr1 register (pop)
	sr0 = dm(ov_stack+5);									// restore primary sr0 register (pop)
	m7 = dm(ov_stack+3);									// restore primary m7 register (pop)
	ax0 = dm(ov_stack+1); 								// restore dmpg2 reg value into ax0 (pop)
	dmpg2 = ax0;													// restore DAG dmpage 2 register
	ay0 = dm(ov_stack+9);									// restore ay0 register (pop)

	flush cache;													// we don't want to cache any of the overlay manager instructions (self-modifying code & cache = bad!!)

	i7 = 0x4000;													// load i7 with address of 1st location of PM run space #2 (PM 0x4000)
	jump (i7) (db);												// jump to location that i7 points to	
	ar = dm(ov_stack+2);									// restore primary ar register (pop)
	i7 = dm(ov_stack+4);									// restore primary i7 register (pop)


DM_Context_Restore:

	//////////////////////////////////////////////////////////////////////////////
	// Here we just perform a context restore for all of the registers that are
	// used by the overlay manager. We have arrived here because the DM overlay
	// module that we requested was already present in one of the DM overlay 'run'
	// spaces.
	//////////////////////////////////////////////////////////////////////////////

	ax0 = dm(ov_stack+8);									// restore iopg register value into ax0 register (pop)
	iopg = ax0;														// restore iopg register
	sr2 = dm(ov_stack+7);									// restore primary sr2 register (pop)
	sr1 = dm(ov_stack+6);									// restore primary sr1 register (pop)
	sr0 = dm(ov_stack+5);									// restore primary sr0 register (pop)
	m7 = dm(ov_stack+3);									// restore primary m7 register (pop)
	ax0 = dm(ov_stack+1); 								// restore dmpg2 reg value into ax0 (pop)
	dmpg2 = ax0;													// restore DAG dmpage 2 register
	ay0 = dm(ov_stack+9);									// restore ay0 register (pop)

	flush cache;													// we don't want to cache any of the overlay manager instructions (self-modifying code & cache = bad!!)

	rts (db);															// return back to main program (or calling function)
	i7 = dm(ov_stack+4);									// restore primary i7 register (pop)
	ar = dm(ov_stack+2);									// restore primary ar register (pop)


