///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Filename:		Overlay Manager DMA Version 1.asm
//							(c) ANALOG DEVICES 2001, gjf
//
// Date:				10/23/2001
// Purpose:			Software Overlay Manager Function
//
// Inputs:			primary ax0 (PLIT_SYMBOL_OVERLAYID), primary ay0 (PLIT_SYMBOL_ADDRESS)
// Outputs:			primary i7 (used to call overlay module for execution of the overlay)
//
// Description:	This progam implements a simple software overlay manager which is uses the DSP's DMA engine to load
//							in the overlay code from external memory. This code module has been left generic enough to allow
//							for simple code reuse and modifications.
//							
// Registers Used:
//
//							primary register set:		ax0, sr2, sr1, sr0, ar, dmpg2, iopg, IJPG (not saved or restored!!!!)
//							primary DAGs:						i7, m7
//							secondary register set:	none
//							secondary DAGs:					none
//
// Additional Information:
//
//							The previous version of this overlay manager utilized a software stack (stored in DM) to perform the
//							context save-restore of ALL of the core registers. This may prove to be impractical for some systems
//							due to the additional overhead of the context save-restore of all of the 2191's registers. This version
//							consumes only a small portion of the 2191's register set which allows for fast context saving-restoring
//							of the registers. Also, since this overlay manager does not use the secondary register set, this code
//							module can be interrupted without any adverse side-effects.
//
//							Currently there is a bug in the linker which places the overlay ID# as the first location (target 'live'
//							address) of the overlay. This affects the count and actual the live address of the overlay. Because of
//							this some additional code has been implemented (labeled 'bug' in the commented section of the code) as
//							a workaround for this bug.
//
// Revision Information:
//
//							rev 0.0 - (10/23/01) -	This revision assumes that ALL calls to the loaded overlay modules are within internal
//																			PM memory. Therefore the IJPG register is NEVER saved or restored. If the user wishes
//																			to execute PM overlays that 'run' from external memory, then care must be taken to
//																			ensure that the IJPG register is properly saved and restored on the overlay stack.
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



#include "2191asm.h"
.section/dm	seg_data;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DMA descriptor blocks below. One descriptor is used for the DMA read channel, the other for the DMA write channel.
// We'll initialize these blocks to zero just as a precaution.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

.var WR_DMA_DESC_BLOCK [5] =	0x0000,		// write DMA configuration register
															0x0000,		// write DMA start page (internal DSP 'live' address of overlay module (ALWAYS assumed as zero!!))
															0x0000,		// write DMA start address (internal DSP 'run' address of overlay module)
															0x0000,		// write DMA word count
															0x0000;		// write DMA next descriptor pointer
															
.var RD_DMA_DESC_BLOCK [5] =  0x0000,		// read DMA configuration register
															0x0000,		// read DMA start page
															0x0000,		// read DMA start address
															0x0000,		// read DMA word count
															0x0000;		// read DMA next descriptor pointer

.var end_dma = 0x0000;									// this variable is used to terminate the DMA (set this variable to a non-zero value for chained DMAs)

.global WR_DMA_DESC_BLOCK;
.global RD_DMA_DESC_BLOCK;
.global end_dma;


.section/pm	seg_code;

.global _OverlayManager;								// address label for the overlay manager function (which is called from the main program)

.extern liveAddresses;									// the 'live' address for the overlay (external memory location)
.extern runAddresses;										// the 'run' address for the overlay (internal DSP memory location)
.extern runWordSize;										// number of internal DSP (PM or DM) memory locations the overlay will occupy
.extern liveWordSize;										// number of external (PM or DM) memory locations the overlay will occupy (same as runWordSize for 2191)

.extern ov_stack;												// software runtime stack (in DM) used for context save-restore by overlay manager function

.extern save_ax0;												// temporary memory locations for saving-restoring of ax0 register
.extern save_ay0;												// temporary memory locations for saving-restoring of ay0 register

.extern ovly_type;											// flag variable used to determine whether the overlay we intend to load in is for PM or DM (=1 for PM)
.extern curr_PM_ovly_ID;								// variable used to store the ID# of the overlay that is to be loaded into internal memory

_OverlayManager:


Context_Save:

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Save registers that the overlay manager will be using onto the overlay manager stack.
// (This will come in handy when we'll be pre-emptively calling an overlay.)
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	dm(ov_stack) = ax0;										// save primary ax0 register onto overlay manager stack (push)
	ax0 = dmpg2;
	dm(ov_stack+1) = ax0;									// save dmpg2 register onto overlay manager stack (push)
	dm(ov_stack+2) = ar;									// save ar register onto overlay manager stack (push)
	dm(ov_stack+3) = m7;									// save m7 register onto overlay manager stack (push)
	dm(ov_stack+4) = i7;									// save i7 register onto overlay manager stack (push)
	dm(ov_stack+5) = sr0;									// save sr0 register onto overlay manager stack (push)
	dm(ov_stack+6) = sr1;									// save sr1 register onto overlay manager stack (push)
	dm(ov_stack+7) = sr2;									// save sr2 register onto overlay manager stack (push)
	ax0 = iopg;
	dm(ov_stack+8) = ax0;									// save iopg register onto overlay manager stack (push)

Find_Overlay:

	dmpg2 = 0;														// switch to the internal memory page #0 (since all of the overlay symbols reside in DMPAGE0)
	ax0 = dm(ov_stack);										// fetch primary ax0 value which contains the overlay ID # (PLIT_SYMBOL_OVERLAYID)
	ar = ax0 - 1;													// find the overlay's tag number
	m7 = ar;															// m7 is the offset needed to get _ov values from the "liveaddresses", "runaddresses", and "pageaddresses" buffers

	dm(curr_PM_ovly_ID) = ar;							// save off the ID# of the overlay (used to check if this overlay is already loaded into memory

	i7 = runAddresses;  									// get the internal address that the overlay is going to reside at runtime
	ax0 = pm(i7 + m7); 										// get the offset of the overlay from the overlay runAddresses buffer
	dm(WR_DMA_DESC_BLOCK+2) = ax0;				// write the overlay 'run' address for MEM DMA WR address reg into its location in MEM DMA WR descriptor block

	dm(ov_stack+19) = ax0;								// gjf 11-05-01: added for ISR to jump to overlay function after DMA fetch is completed

	// We don't need to update the DMA WR page register since it's initialized to zero in the descriptor block by default
	// and also because for our example ALL of the overlay regions run from internal PM (i.e. ijpg=0).
	// The code for implementing this will be provided below for additional modification of this example by the user.
	

Parse_Live_Address:

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Since the live address is a 24-bit value we have to parse this 24-bit value
// into the 8-bit page and 16-bit address offset fields to update the DMPGx and DAG registers accordingly
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	i7 = runWordSize;											// i7 points to head of the 'runWordSize' buffer
	ax0 = dm(i7 + m7);  									// get the word count of the desired overlay (from m7 offset)
	ar = ax0;															// ax0 equals the word count of the desired overlay
	ar = ar - 1;													// subtract 1 from the word count (bug)
	dm(WR_DMA_DESC_BLOCK+3) = ar;					// write word count for MEM DMA word count into its location in the DMA WR descriptor block
	dm(RD_DMA_DESC_BLOCK+3) = ar;					// read word count for MEM DMA word count into its location in the DMA RD descriptor block

	i7 = liveAddresses;										// i7 points to head of liveAddresses buffer
	ar = pm(i7 + m7);											// get upper 16 MSBs of 24-bit address constant label of the desired overlay (from m7 offset)
	sr = lshift ar by 0x8 (lo);						// shift value 8 bits left to place page value in SR1
	ar = px;															// get 8 LSBs of 24-bit address constant label
	sr = sr or lshift ar by 0x0 (lo);			// OR the 8 LSBs with the parsed 16 LSBs from prior shift instruction
	i7 = sr0;															// i7 points to the live address of the desired overlay
	ar = i7;															// load target address into ar (bug)
	ar = ar + 1;													// increment overlay target address by 1 because overlay ID is loaded as the 1st location of the overlay (bug)

	dm(RD_DMA_DESC_BLOCK+2) = ar;					// write overlay 'live' address for MEM DMA RD address reg into its location in MEM DMA RD descriptor block
	dm(RD_DMA_DESC_BLOCK+1) = sr1;				// write overlay 'live' page for MEM DMA RD page reg into its location in MEM DMA RD descriptor block

	ax0 = 0x800f;													// configure MEM DMA WR Config Reg for 24-bit writes (to int mem), enable DMA, DMA descriptor ownership, interrupt on completion
	dm(WR_DMA_DESC_BLOCK) = ax0;					// write this value for MEM DMA WR Config Reg into its location in MEM DMA WR descriptor block
	ax0 = 0x800d;													// configure MEM DMA RD Config Reg for 24-bit reads (from ext mem), enable DMA, DMA descriptor ownership, interrupt on completion
	dm(RD_DMA_DESC_BLOCK) = ax0;					// write this value for MEM DMA RD Config Reg into its location in MEM DMA RD descriptor block
	
	ax0 = end_dma;												// ax0 equals address of 'end_dma' variable
	dm(WR_DMA_DESC_BLOCK+4) = ax0;				// this writes the value of 'end_dma' to end the DMA chaining sequence (=0)
	dm(RD_DMA_DESC_BLOCK+4) = ax0;				// this writes the value of 'end_dma' to end the DMA chaining sequence (=0)
	

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Here's where we finally kick off the DMA and fetch the overlay module !!!!!!!!!
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	iopg =  Memory_DMA_Controller_Page;		// switch to MEM DMA IO page (0x01)

	ax0 = WR_DMA_DESC_BLOCK;							// ax0 points to the head of the write DMA descriptor block (linked-list)
	io(DMACW_CP) = ax0;										// load the address of the DMA write channel descriptor block into write chain pointer register
	ax0 = RD_DMA_DESC_BLOCK;							// ax0 points to the head of the write DMA descriptor block (linked-list)
	io(DMACR_CP) = ax0;										// load the address of the DMA read channel descriptor block into read chain pointer register

	ax0 = 0x1;
	io(DMACW_CPR) = ax0;									// set the descriptor ready bit for the DMA write channel
	io(DMACR_CPR) = ax0;									// set the descriptor ready bit for the DMA read channel

	io(DMACW_CFG) = ax0;									// enable the write DMA channel
	io(DMACR_CFG) = ax0;									// enable the read DMA channel

	ena int;															// enable global interrupts (just in case???)

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////	
// gjf 11-07-01: added the following code here for the overlay manager to have the idle instruction here for the MEM DMA interrupt
	idle;

Overlay_Loaded:

	nop;		// this guy here just for debugging purposes only

// gjf 11-07-01: end of added code
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////	


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// At this point we're done with loading in the overlay region. Now all we have to do is restore the scratch registers used
// by the overlay manager and jump to the overlay function to begin execution.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Context_Restore:

	ax0 = dm(ov_stack+8);									// restore iopg register value into ax0 register (pop)
	iopg = ax0;														// restore iopg register
	sr2 = dm(ov_stack+7);									// restore primary sr2 register (pop)
	sr1 = dm(ov_stack+6);									// restore primary sr1 register (pop)
	sr0 = dm(ov_stack+5);									// restore primary sr0 register (pop)
	
	// gjf 11-07-01: removed the line below
	//i7 = dm(ov_stack+4);									// restore primary i7 register (pop)
	
	m7 = dm(ov_stack+3);									// restore primary m7 register (pop)
	//ar = dm(ov_stack+2);									// restore primary ar register (pop)

	ax0 = dm(ov_stack+1); 								// restore dmpg2 reg value into ax0 (pop)
	dmpg2 = ax0;													// restore DAG dmpage 2 register (this instruction placed here because we don't want to jump to incorrect page!!)

// we don't need to restore ax0 here because it's saved by the calling function already

	flush cache;													// we don't want to cache any of the overlay manager instructions (self-modifying code & cache = bad!!)

Jump_To_Caller:

	//rts (db);															// return back to main program (or calling function)
	//ax0 = dm(ov_stack+1); 								// restore dmpg2 reg value into ax0 (pop)
	//dmpg2 = ax0;													// restore DAG dmpage 2 register (this instruction placed here because we don't want to jump to incorrect page!!)

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////	
// gjf 11-07-01: added the following code here for the overlay manager

	i7 = dm(ov_stack+19);	// gjf 11-07-01: i7 = overlay run address
	jump (i7) (db);	// gjf 11-07-01: jump to overlay function
	i7 = dm(ov_stack+4);
	ar = dm(ov_stack+2);

